
from dataclasses import dataclass, field
from abc import abstractmethod

import itertools
import pathlib

from collections.abc import Sequence
from typing import Union as U
from typing import Any, Generator, Iterable, Optional

INDENT = '    '

#══════════════════════════════════════════════════════════════════════════════
@dataclass
class ElementBase:
    def __add__(self, other):
        return Document([self, other])
    #─────────────────────────────────────────────────────────────────────────
    @abstractmethod
    def __str__(self) -> str:
        pass
#══════════════════════════════════════════════════════════════════════════════
def list_diff(list1, list2):
    return [item for item in list1 if item not in set(list2)]
#══════════════════════════════════════════════════════════════════════════════
def collect_references_iter(items: Iterable) -> list:
    '''Doing ordered set union thing

    Parameters
    ----------
    items : Iterable
        items to be checked

    Returns
    -------
    list
        list of unique items
    '''
    output: list[Link] = []
    for item in items:
        if hasattr(item, '_collect_references'):
            output += list_diff(item._collect_references(), output)
    return output
#══════════════════════════════════════════════════════════════════════════════
notations = {'bold': '**',
             'italic': '*',
             'strikethrough': '~~',
             'subscript': '~',
             'superscript': '^',
             'emphasis': '=='}
@dataclass
class StylisedText(ElementBase):
    '''Inline text

    Parameters
    ----------
    text : has method str 
        text to be containes
    style: set
        style of the text, options are: bold, italic, strikethrough, subscript, superscript, emphasis

    Raises
    ------
    ValueError
        for invalid style attributes
    '''
    text: str
    style: set[str] = field(default_factory = set)
    def __post_init__(self):
        if incorrect_substyles := [substyle for substyle in self.style
                                   if substyle not in notations]:
            raise ValueError(f'Style options {incorrect_substyles} invalid')

        if 'subscript' and 'superscipt' in self.style:
            raise ValueError('Text cannot be both superscript and subscript')
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        text = str(self.text)
        for substyle in self.style:
            marker = notations[substyle]
            text = f'{marker}{text}{marker}'
        return text
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Address(ElementBase):
    text: str
    def __str__(self) -> str:
        return f'<{self.text}>'
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Monospace(ElementBase):
    text: str
    def __str__(self) -> str:
        return f'`{self.text}`'
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Paragraph(ElementBase):
    content: list = field(default_factory = list)
    #─────────────────────────────────────────────────────────────────────────
    def _collect_references(self) -> list:
        return collect_references_iter(self.content)
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        return ''.join(str(item) for item in self.content)
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Link(ElementBase):
    text: U[str, ElementBase]
    url: U[str, pathlib.Path]
    title: Optional[str] = None
    _index: int = 0
    _hash: Optional[int] = None
    #─────────────────────────────────────────────────────────────────────────
    def _collect_references(self) -> list:
        if self.title is not None:
            return [self]
        return []
    #─────────────────────────────────────────────────────────────────────────
    def __hash__(self) -> int:
        if self._hash is None:
            return hash(str(self.text)) + hash(str(self.url)) + hash(str(self.title))
        return self._hash
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        if self._index:
            return f'[{self.text}][{self._index}]'
        return f'[{self.text}]({self.url})'
#══════════════════════════════════════════════════════════════════════════════
listingprefixes = {'unordered': (itertools.repeat('- '), 2),
                   'ordered': ((f'{n}. ' for n in itertools.count(start = 1, step = 1)), 3),
                   'checkbox': (itertools.repeat('- [ ] '), 6),
                   'checkbox': (itertools.repeat('- [x] '), 6),
                   'definition': (itertools.repeat(': '), 2)}
@dataclass
class Listing(ElementBase):
    listingtype: str
    content: list
    #─────────────────────────────────────────────────────────────────────────
    def __post_init__(self):
        if self.listingtype not in listingprefixes:
            raise ValueError(f'{self.listingtype} not in listingprefixes')
        if isinstance(self.content[0], Listing):
            raise ValueError(f'First item cannot be sublist')
    #─────────────────────────────────────────────────────────────────────────
    def _collect_references(self) -> list:
        return collect_references_iter(self.content)
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        prefix, prefix_length = listingprefixes[self.listingtype]
        output = ''
        for item in self.content:
            if isinstance(item, tuple):
                output += next(prefix) + str(item[0]) + '\n'
                output += INDENT + str(item[1]).replace('\n', '\n'+ INDENT
                                                     ).removesuffix(INDENT) + '\n'
            else:
                output += next(prefix) + str(item).replace('\n', '\n'+ ' '* prefix_length) + '\n'
        return output[:-1]
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class InlineMath(ElementBase):
    text: U[str, ElementBase]
    def __str__(self) -> str:
        return f'${self.text}$'
#══════════════════════════════════════════════════════════════════════════════
def padd(items: Iterable[str], widths: Iterable[int]
         ) -> Generator[str, None, None]:
    for item, width in zip(items, widths):
        item += ((width - len(item))//2) * ' '
        yield (width - len(item)) * ' ' + item
@dataclass
class Table(ElementBase):
    header: Iterable
    alignment: list[str]
    content: list[Sequence] = field(default_factory = list)
    compact: bool = False
    #─────────────────────────────────────────────────────────────────────────
    def _collect_references(self):
        output = collect_references_iter(self.header)
        for row in self.content:
            output += list_diff(collect_references_iter(row), output)
        return output
    #─────────────────────────────────────────────────────────────────────────
    def append(self, row: Sequence) -> None:
        if hasattr(row, '__iter__'):
            self.content.append(row)
        else:
            raise TypeError(f"'{type(row)}' object is not iterable")
    #─────────────────────────────────────────────────────────────────────────
    @staticmethod
    def _str_row_sparse(row: Iterable):
        return '| ' + ' | '.join(row) + ' |\n'
    #─────────────────────────────────────────────────────────────────────────
    @staticmethod
    def _str_row_compact(row: Iterable):
        return '|' + '|'.join(row) + '|\n'
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        header = [str(item) for item in self.header]
        headerlen = len(header)
        max_rowlen = headerlen

        for row in self.content:
            if len(row) > max_rowlen:
                max_rowlen = len(row)
        alignment = self.alignment + ['left'] * (max_rowlen - len(self.alignment))
        if self.compact:
            output = self._str_row_compact(header) + '|\n'
            # Alignments
            for i, item in enumerate(alignment):
                if item == 'left':
                    alignment[i] = ' :-- '
                    continue
                if item == 'center':
                    alignment[i] = ' :-: '
                    continue
                if item == 'right':
                    alignment[i] = ' --: '
            output += self._str_row_compact(item for item in alignment)
            for row in self.content:
                output += self._str_row_compact(str(item) for item in row)
            return output[:-1] # Removing the last \n

        # pretty
        max_widths = [max(len(item), 3) for item in header] + (max_rowlen - headerlen) * [3]
        content = [[str(item) for item in row] for row in self.content]
        for row in content:
            for i, item in enumerate(row):
                itemlen = len(item)
                if itemlen > max_widths[i]:
                    max_widths[i] = itemlen
        output = self._str_row_sparse(padd(header, max_widths))
        # Alignments
        for i, item in enumerate(alignment):
            if item == 'left':
                alignment[i] = ' :'+ (max_widths[i] - 1) * '-'
                continue
            if item == 'center':
                alignment[i] = ' :'+ (max_widths[i] - 2) * '-' + ':'
                continue
            if item == 'right':
                alignment[i] = (max_widths[i] - 1) * '-' + ':'
        output += self._str_row_sparse(item for item in alignment)
        for row in content:
            output += self._str_row_sparse(padd(row, max_widths))
        return output[:-1]
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class QuoteBlock(ElementBase):
    content: ElementBase
    #─────────────────────────────────────────────────────────────────────────
    def _collect_references(self) -> list:
        if hasattr(self.content, '_collect_references'):
            return self.content._collect_references()
        return []
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        content = '> ' + str(self.content).replace('\n', '\n> ')
        return content
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Heading(ElementBase):
    level: int
    text: str
    include_in_TOC: bool = True
    def __str__(self) -> str:
        return f'{self.level * "#"} {self.text}' + ('' if self.include_in_TOC
                                                    else " <!-- omit in toc -->")
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Block(ElementBase):
    language: str
    content: str
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self) -> str:
        return f'```{self.language}\n{self.content}\n```'
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Image(ElementBase):
    path: U[str, pathlib.Path]
    alt_text: str = ''
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self):
        return f'![{self.alt_text}]({self.path})'
#══════════════════════════════════════════════════════════════════════════════
@dataclass
class Document:
    content: list = field(default_factory = list)
    #─────────────────────────────────────────────────────────────────────────
    def __iadd__(self, item):
        self.content.append(item)
        return self
    #─────────────────────────────────────────────────────────────────────────
    def __add__(self, item):
        if isinstance(item, Document):
            self.content + item.content
            return self
        else:
            self.content.append(item)
            return self
    #─────────────────────────────────────────────────────────────────────────
    def __str__(self)  -> str:
        # Collecting references
        references = collect_references_iter(self.content)
        for index, link in enumerate(references, start = 1):
            link._index = index
        reftext = '\n'+'\n'.join(f'[{link._index}]: <{link.url}> "{link.title}"'
                                 for link in references) + '\n'
        return '\n\n'.join(str(item) for item in self.content) + '\n' + reftext
    #─────────────────────────────────────────────────────────────────────────
    def to_file(self,
                filepath: pathlib.Path = pathlib.Path.cwd() / 'document.md'
                ) -> None:
        with open(filepath, 'w+', encoding  = 'utf8') as f:
            f.write(str(self))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#%%═════════════════════════════════════════════════════════════════════
# IMPORT


import pathlib
from setuptools import find_packages, setup
import sys
if '--tests' in sys.argv or '--print' in sys.argv:
    import colorama as col
    RESET = col.Style.RESET_ALL
    BLACK = col.Fore.BLACK
    BLUE = col.Fore.BLUE
    CYAN = col.Fore.CYAN
    GREEN = col.Fore.GREEN
    MAGENTA = col.Fore.MAGENTA
    RED = col.Fore.RED
    YELLOW = col.Fore.YELLOW
    WHITE = col.Fore.WHITE
    WHITE_BG = col.Back.WHITE
#%%═════════════════════════════════════════════════════════════════════
# SETUP GLOBALS
BASE_DIR = pathlib.Path(__file__).parent
PACKAGE_NAME = 'YAMDOG'
PYTHON_VERSION = '>=3.9'
PATH_LICENCE = next(BASE_DIR.glob('LICENSE*'))
PATH_SCR = BASE_DIR / 'src'
PATH_INIT = next(PATH_SCR.rglob('__init__.py'))
IMPORT_NAME = PATH_INIT.parent.name
PATH_README = next(BASE_DIR.glob('README*'))
#%%═════════════════════════════════════════════════════════════════════
# Run tests first
if '--tests' in sys.argv:
    import tests
    print('Running typing checks')
    typing_test_result = tests.typing(shell = False)
    failed = not typing_test_result[0].startswith('Success')
    failed |= bool(typing_test_result[1])
    print(f'{RED if failed else GREEN}{typing_test_result[0]}{RESET}')
    if typing_test_result[1]:
        print(typing_test_result[1])

    print('Running unit tests')
    unit_test_result = tests.unittests(verbosity = 1)
    failed |= bool(unit_test_result.errors)
    failed |= bool(unit_test_result.failures)
    if failed:
        raise Exception('Tests did not pass, read above')
    sys.argv.pop(sys.argv.index('--tests'))
#%%═════════════════════════════════════════════════════════════════════
# SETUP FUNCTIONS
def header(text: str, linechar = '─', endchar = '┐', headerwidth  =  60):
    titlewidth = headerwidth // 2
    textlen = len(text)
    l_len = ((titlewidth - textlen) // 2 - 1)
    lpad = linechar*l_len
    rpad = f'{(headerwidth - l_len - textlen - 3)*linechar}'
    return f'{lpad} {GREEN}{text}{RESET} {rpad}{endchar}'
#───────────────────────────────────────────────────────────────────────
# For classifiers
def c(*args):
    out = f'{args[0]} :: {args[1]}'
    for arg in args[2:]:
        out += f' :: {arg}'
    return out
#───────────────────────────────────────────────────────────────────────
def cset(key, *values):
    out = []
    if isinstance(key, str):
        key = (key, )
    for value in values:
        if isinstance(value, tuple):
            out.append(c(*key, *value))
        else:
            out.append(c(*key, value))
    return out
#%%═════════════════════════════════════════════════════════════════════
# SETUP INFO
if '--print' in sys.argv:
    print(f'\n{header("Starting packaging setup", "=", "=")}\n')
setup_info = {}
# Getting package name
setup_info['name'] = PACKAGE_NAME
#───────────────────────────────────────────────────────────────────────
# Version
with open(PATH_INIT, 'r', encoding = 'utf8') as f:
    while not (line := f.readline().lstrip()).startswith('__version__'):
        pass
    setup_info['version'] = line.split('=')[-1].strip(" '")
#───────────────────────────────────────────────────────────────────────
# Licence
with open(PATH_LICENCE, 'r', encoding = 'utf8') as f:
    LICENSE_NAME = f'{f.readline().strip().split()[0]}'
setup_info['license'] = LICENSE_NAME
#───────────────────────────────────────────────────────────────────────
# Author
setup_info['author'] = 'Limespy'
#───────────────────────────────────────────────────────────────────────
# Author Email
setup_info['author_email'] = ''
#───────────────────────────────────────────────────────────────────────
# URL
setup_info['url'] = f'https://github.com/{setup_info["author"]}/{setup_info["name"]}'
GITHUB_MAIN_URL = f'{setup_info["url"]}/blob/main/'
#───────────────────────────────────────────────────────────────────────
# Description
with open(PATH_README, 'r', encoding = 'utf8') as f:
    # The short description is in the README after badges
    while (description := f.readline().lstrip(' ')).startswith(('#', '\n', '[')):
        pass
    while not (line := f.readline().lstrip(' ')).startswith('\n'):
        description += line

setup_info['description'] = description[:-1] # Removing trailing linebreak
#───────────────────────────────────────────────────────────────────────
# Long Description
with open(PATH_README, 'r', encoding = 'utf8') as f:
    setup_info['long_description'] = f.read().replace('./', GITHUB_MAIN_URL)
if PATH_README.suffix == '.md':
    setup_info['long_description_content_type'] = 'text/markdown'
elif PATH_README.suffix != '.rst':
    raise TypeError(f'README file type not recognised: {PATH_README}')
#───────────────────────────────────────────────────────────────────────
# packages
setup_info['packages']  = find_packages('src')
#───────────────────────────────────────────────────────────────────────
# Packages Dir
setup_info['package_dir']  = {'': 'src'}
#───────────────────────────────────────────────────────────────────────
# Py Modules
setup_info['py_modules'] = [path.stem for path in PATH_SCR.rglob('*.py')]
#───────────────────────────────────────────────────────────────────────
# Include  Package Data
setup_info['include_package_data'] = True
#───────────────────────────────────────────────────────────────────────
# Classifiers
# complete classifier list:
#   http://pypi.python.org/pypi?%3Aaction=list_classifiers
setup_info['classifiers']   = [
    c('Development Status', '3 - Alpha'),
    c('License', 'OSI Approved', 'MIT License'),
    *cset('Operating System', 'Unix', 'POSIX', ('Microsoft', 'Windows')),
    *cset(('Programming Language', 'Python'),
          '3', ('3', 'Only'), '3.9', '3.10', '3.11')]
#───────────────────────────────────────────────────────────────────────
# Project URLs
setup_info['project_urls'] = {
    'Changelog': f'{GITHUB_MAIN_URL}{PATH_README.name}#Changelog',
    'Issue Tracker': f'{setup_info["url"]}/issues'}
#───────────────────────────────────────────────────────────────────────
# Keywords
setup_info['keywords'] = ['markdown']
#───────────────────────────────────────────────────────────────────────
# Python requires
setup_info['python_requires']  = PYTHON_VERSION
#───────────────────────────────────────────────────────────────────────
# Install requires
with open(BASE_DIR / 'dependencies.txt', encoding = 'utf8') as f:
    setup_info['install_requires'] = [line.rstrip() for line in f.readlines()]
#───────────────────────────────────────────────────────────────────────
# Extras require
with open(BASE_DIR / 'dependencies_dev.txt', encoding = 'utf8') as f:
    setup_info['extras_require'] = {'dev':
                                    [line.rstrip() for line in f.readlines()]}
#───────────────────────────────────────────────────────────────────────
# Entry points
setup_info['entry_points'] = {'console_scripts':
    [f'{IMPORT_NAME} = {IMPORT_NAME}.CLI:main',]}
#%%═════════════════════════════════════════════════════════════════════
# PRINTING SETUP INFO
if '--print' in sys.argv:
    for key, value in setup_info.items():
        print(f'\n{header(key)}\n')
        if isinstance(value, list):
            print('[', end = '')
            if value:
                print(value.pop(0), end = '')
                for item in value:
                    print(f',\n {item}', end = '')
            print(']')
        elif isinstance(value, dict):
            print('{', end = '')
            if value:
                items = iter(value.items())
                key2, value2 = next(items)
                print(f'{key2}: {value2}', end = '')
                for key2, value2 in items:
                    print(f',\n {key2}: {value2}', end = '')
            print('}')
        else:
            print(value)
    print(f'\n{header("Calling setup", "=", "=")}\n')
    sys.argv.pop(sys.argv.index('--print'))
#%%═════════════════════════════════════════════════════════════════════
# RUNNING THE SETUP
setup(**setup_info)
